//
// $Id: input.cc,v 1.6 1999/08/08 09:07:42 voeckler Exp $
//
// Author:  Jens-S. Vckler <voeckler@rvs.uni-hannover.de>
//
// File:    input.cc
//          Sat Jul  4 1998
//
// (c) 1998 Lehrgebiet Rechnernetze und Verteilte Systeme
//          Universitt Hannover, Germany
//
// Permission to use, copy, modify, distribute, and sell this software
// and its documentation for any purpose is hereby granted without fee,
// provided that (i) the above copyright notices and this permission
// notice appear in all copies of the software and related documentation,
// and (ii) the names of the Lehrgebiet Rechnernetze und Verteilte
// Systeme and the University of Hannover may not be used in any
// advertising or publicity relating to the software without the
// specific, prior written permission of Lehrgebiet Rechnernetze und
// Verteilte Systeme and the University of Hannover.
//
// THE SOFTWARE IS PROVIDED "AS-IS" AND WITHOUT WARRANTY OF ANY KIND,
// EXPRESS, IMPLIED OR OTHERWISE, INCLUDING WITHOUT LIMITATION, ANY
// WARRANTY OF MERCHANTABILITY OR FITNESS FOR A PARTICULAR PURPOSE.
//
// IN NO EVENT SHALL THE LEHRGEBIET RECHNERNETZE UND VERTEILTE SYSTEME OR
// THE UNIVERSITY OF HANNOVER BE LIABLE FOR ANY SPECIAL, INCIDENTAL,
// INDIRECT OR CONSEQUENTIAL DAMAGES OF ANY KIND, OR ANY DAMAGES
// WHATSOEVER RESULTING FROM LOSS OF USE, DATA OR PROFITS, WHETHER OR NOT
// ADVISED OF THE POSSIBILITY OF DAMAGE, AND ON ANY THEORY OF LIABILITY,
// ARISING OUT OF OR IN CONNECTION WITH THE USE OR PERFORMANCE OF THIS
// SOFTWARE.
//
// $Log: input.cc,v $
// Revision 1.6  1999/08/08 09:07:42  voeckler
// added compile-time configurable libz and libbz2 support.
//
// Revision 1.5  1999/08/05 21:10:35  voeckler
// reworked number parsing methods to be more efficient. Added new methods
// for the MyXIntXX types. Improved PlainInput parsing. Added bzip2 input.
//
// Revision 1.4  1999/01/18 08:58:53  voeckler
// snapshot.
//
// Revision 1.3  1998/08/06 20:02:50  voeckler
// major performance issue improvements, restructuring.
//
// Revision 1.2  1998/07/23 18:14:41  voeckler
// replaced rational() with an improved scanner.
//
// Revision 1.1  1998/07/05 15:37:24  voeckler
// Initial revision
//
//
#if defined(__GNUC__) || defined(__GNUG__)
#pragma implementation
#endif

#include "input.hh"
#include "ctype.hh"
#include <errno.h>
#include <memory.h>

#include <assert.h>
#include <ctype.h>
#include <stdlib.h>
#include <string.h>
#include <unistd.h>
#include <stdio.h>
#include <float.h>
#include <math.h>

static const char* RCS_ID =
"$Id: input.cc,v 1.6 1999/08/08 09:07:42 voeckler Exp $";

#define SGET \
  int count = 0; \
  bool negate = false; \
  \
  number = 0; \
  skiplws(); \
  int ch = peek(); \
  if ( ch == '+' || ch == '-' ) { \
    negate = ( ch == '-' ); \
    get(); /* advance cursor */ \
    count++; \
  } \
  \
  /* PRECONDITION: chars 0..-2 are not true on ISDIGIT() */ \
  while ( ISDIGIT(ch=peek()) ) { \
    number = number*10 + ( ch - '0' ); \
    count++; \
    get(); \
  } \
  if ( negate ) number = -number; \
  return count; 

#define UGET \
  int ch, count = 0; \
  number = 0ul; \
  skiplws(); \
  /* PRECONDITION: chars 0..-2 are not true on ISDIGIT() */ \
  while ( ISDIGIT(ch=peek()) ) { \
    number = number * 10 + ( ch - '0' ); \
    count++; \
    get(); \
  } \
  return count;

int 
BaseInput::integer( MyUInt64& number )
  // purpose: reads an integer from the input
  // paramtr: number (OUT): \s*[0-9]* from current position; 0 for text
  // returns: number of characters viewed, not including LWS, may be 0
  // warning: based on get() and peek()
{
  UGET
}

int 
BaseInput::integer( MySInt64& number )
  // purpose: reads an integer from the input
  // paramtr: number (OUT): \s*[+-]?[0-9]* from current position; 0 for text
  // returns: number of characters viewed, not including LWS, may be 0
  // warning: based on get() and peek()
{
  SGET
}

int 
BaseInput::integer( MyUInt32& number )
  // purpose: reads an integer from the input
  // paramtr: number (OUT): \s*[0-9]* from current position; 0 for text
  // returns: number of characters viewed, not including LWS, may be 0
  // warning: based on get() and peek()
{
  UGET
}

int 
BaseInput::integer( MySInt32& number )
  // purpose: reads an integer from the input
  // paramtr: number (OUT): \s*[+-]?[0-9]* from current position; 0 for text
  // returns: number of characters viewed, not including LWS, may be 0
  // warning: based on get() and peek()
{
  SGET
}

int
BaseInput::rational( double& stamp )
  // purpose: reads a Squid UTC timestamp.millis from the input
  // paramtr: stamp (OUT): \s*[+-]?(\d+)|(\d*[.,]\d+), may be 0.0
  // returns: number of characters viewed, not including LWS, may be 0
  // warning: this is *not* be the complete C float set.
  // warning: based on get() and peek()
{
  static const double exponent[20] = {
     1E00, 1E-01, 1E-02, 1E-03, 1E-04, 1E-05, 1E-06, 1E-07, 1E-08, 1E-09,
    1E-10, 1E-11, 1E-12, 1E-13, 1E-14, 1E-15, 1E-16, 1E-17, 1E-18, 1E-19
  };

  long prae, post;
  int c1 = integer( prae );
  if ( peek() != '.' ) {
    stamp = prae;
    return c1;
  } else {
    get();
    c1++;
  }

  int c2 = integer( post );
  stamp = prae + post / exponent[c2];
  return c1+c2;
}

int
BaseInput::get( char* destination, size_t maxsize )
  // purpose: read a string with maximum length into a buffer
  // paramtr: destination (OUT): area to write string into
  //          maxsize (IN): maximum size of previous area
  // returns: number of characters transferred, may be 0
  //          destination[0] is guaranteed to be 0 for return value 0.
  // warning: based on get() and peek()
{
  int ch;
  char* s = destination;
  char* e = s + maxsize-1;

  skiplws();
  while ( (ch=peek()) >= 0 && !ISSPACE(ch) && s < e ) *s++ = get();
  *s = 0;
  if ( !ISSPACE(ch) ) skiptolws(); // handle overrun
  return s-destination;
}

void 
BaseInput::skiptolws()
  // purpose: skip forward to next linear whitespace [ \t]
{
  for ( int ch; (ch=peek()) >= 0 && !ISBLANK(ch); ch=get() ) ;
}

void 
BaseInput::skiplws()
  // purpose: skip forward during linear whitespaces [ \t]
{
  for ( int ch; (ch=peek()) > 0 && ISBLANK(ch); ch=get() ) ;
}

void 
BaseInput::skippasteoln()
  // purpose: skip forward *over* the next '\n'
  // sidekck: increments line number lineno.
{
  int ch = get();
  while ( ch > 0 && ch != '\n' ) ch=get();
  if ( ch=='\n' ) lineno++;
}




PlainInput::PlainInput( int _fd, int buffersize )
  :fd(_fd),
   bsize(buffersize),
   rsize(1),
   buffer(new PlainInput::uchar[bsize])
  // purpose: ctor
  // paramtr: _fd (IN): file opened for reading
  //          buffersize (IN): size of input buffer to allocate
{
  assert(buffer != 0);
  lstart = start = end = buffer;
}

bool
PlainInput::eof()
{
  // if buffers are empty, force reading of next chunk
  if ( start >= end && rsize > 0 ) peek();
  // an error is also an EOF condition...
  return ((start >= end) && (rsize <= 0)); 
}

int
PlainInput::peek()
  // purpose: obtain the next character to be returned, w/o getting it.
  //          actually handles the read() calls, too.
  // returns: next character, or -1 for error, or -2 for EOF.
{
  // don't read if error or EOF
  if ( rsize <= 0 ) return ((-2)-rsize);

  if ( start == end ) {
    do {
      rsize = read( fd, (char*) buffer, bsize );
    } while ( rsize == -1 && errno == EINTR );
    
    if ( rsize > 0 ) {
      // position end one past buffer
      start = buffer;
      end = buffer + rsize;
    } else {
      // eof or error
      return ((-2)-rsize);
    }
  }

  return *start;
}
  
int
PlainInput::get( char* destination, size_t maxsize )
  // purpose: read a string with maximum length into a buffer
  // paramtr: destination (OUT): area to write string into
  //          maxsize (IN): maximum size of previous area
  // returns: number of characters transferred, may be 0
  //          destination[0] is guaranteed to be 0 for return value 0.
  // this is a slightly more efficient refinement
{
  int ch;
  char* s = destination;
  char* e = s + maxsize-1;

  // though this looks more complicated, it is actually much faster,
  // since it saves repeated calls to peek().
  skiplws();
  while ( (ch=peek()) >= 0 && !ISSPACE(ch) && s < e ) {
    *s++ = *start++;
    while ( start < end && !ISSPACE(*start) && s < e ) *s++ = *start++;
  }
  *s = 0;
  if ( !ISSPACE(ch) ) skiptolws(); // handle overrun
  return s-destination;
}

void 
PlainInput::skiptolws()
  // purpose: skip forward to next linear whitespace [ \t]
  // this is a slightly more efficient refinement
{
  while ( start < end && !ISSPACE(*start) )
    if ( ++start >= end )
      // peek() will reset 'start', if there is data available
      if ( peek() < 0 ) break;
}

void 
PlainInput::skiplws()
  // purpose: skip forward during linear whitespaces [ \t]
  // this is a slightly more efficient refinement
{
  while ( start < end && ISSPACE(*start) )
    if ( ++start >= end )
      // peek() will reset 'start', if there is data available
      if ( peek() < 0 ) break;
}

void 
PlainInput::skippasteoln()
  // purpose: skip forward *over* the next '\n'
  // sidekck: increments line number lineno and updates lstart.
{
  int ch;

  do {
    ch = *start++;
    if ( start >= end && peek() < 0 ) break;
  } while ( ch != '\n' );

  if ( ch=='\n' ) {
    lineno++;
    lstart = start;
  }
}

int 
PlainInput::position() const
  // purpose: return position from start of line
  // warning: unreliable for very long lines
{
  return ( lstart <= start ? start-lstart : (end-lstart) + (start-buffer) );
}

#undef SGET
#define SGET \
  int count = 0; \
  bool negate = false; \
  \
  number = 0; \
  skiplws(); \
  if ( *start == '+' || *start == '-' ) { \
    negate = ( *start == '-' ); \
    if ( ++start >= end ) peek(); \
    count++; \
  } \
  \
  /* PRECONDITION: chars 0..-2 are not true on ISDIGIT() */ \
  while ( ISDIGIT(peek()) ) { \
    count++; \
    number = number*10 + ( *start++ - '0' ); \
  } \
  \
  if ( negate ) number = -number; \
  return count;

#undef UGET
#define UGET \
  int count = 0; \
  \
  number = 0; \
  skiplws(); \
  /* PRECONDITION: chars 0..-2 are not true on ISDIGIT() */ \
  while ( ISDIGIT(peek()) ) { \
    count++; \
    number = number*10 + ( *start++ - '0' ); \
  } \
  \
  return count;

int 
PlainInput::integer( MyUInt64& number )
  // purpose: reads an integer from the input
  // paramtr: number (OUT): \s*[0-9]* from current position; 0 for text
  // returns: number of characters viewed, not including LWS, may be 0
  // this is a slightly more efficient refinement
{
  UGET
}

int 
PlainInput::integer( MySInt64& number )
  // purpose: reads an integer from the input
  // paramtr: number (OUT): \s*[+-]?[0-9]* from current position; 0 for text
  // returns: number of characters viewed, not including LWS, may be 0
  // this is a slightly more efficient refinement
{
  SGET
}

int 
PlainInput::integer( MyUInt32& number )
  // purpose: reads an integer from the input
  // paramtr: number (OUT): \s*[0-9]* from current position; 0 for text
  // returns: number of characters viewed, not including LWS, may be 0
  // this is a slightly more efficient refinement
{
  UGET
}

int 
PlainInput::integer( MySInt32& number )
  // purpose: reads an integer from the input
  // paramtr: number (OUT): \s*[+-]?[0-9]* from current position; 0 for text
  // returns: number of characters viewed, not including LWS, may be 0
  // this is a slightly more efficient refinement
{
  SGET
}

#ifdef USE_LIBZ

GZipInput::GZipInput( int _fd, int buffersize )
  :PlainInput(_fd,buffersize),
   // gzclose will close _fd, if not dup()ed
   zipfile( gzdopen( dup(_fd), "rb" ) )
{
  assert( zipfile != NULL );
}

GZipInput::~GZipInput()
{
  gzclose(zipfile);
}

int
GZipInput::peek()
  // purpose: obtain the next character to be returned, w/o getting it.
  //          actually handles the read() calls, too.
  // returns: next character, or -1 for error, or -2 for EOF.
{
  // don't read if error or EOF
  if ( rsize <= 0 ) return ((-2)-rsize);

  if ( start == end ) {
    do {
      rsize = gzread( zipfile, buffer, bsize );
    } while ( rsize == -1 && errno == EINTR );
    
    if ( rsize > 0 ) {
      // position end one past buffer
      start = buffer;
      end = buffer + rsize;
    } else {
      // eof or error
      return ((-2)-rsize);
    }
  }

  return ((unsigned char)(*start));
}

#endif // USE_LIBZ

#ifdef USE_LIBBZ2

BZip2Input::BZip2Input( int _fd, int buffersize )
  :PlainInput(_fd,buffersize),
   // bzclose might close _fd, if not dup()ed
   zipfile( bzdopen( dup(_fd), "rb" ) )
{
  assert( zipfile != NULL );
}

BZip2Input::~BZip2Input()
{
  bzclose(zipfile);
}

int
BZip2Input::peek()
  // purpose: obtain the next character to be returned, w/o getting it.
  //          actually handles the read() calls, too.
  // returns: next character, or -1 for error, or -2 for EOF.
{
  // don't read if error or EOF
  if ( rsize <= 0 ) return ((-2)-rsize);

  if ( start == end ) {
    do {
      rsize = bzread( zipfile, buffer, bsize );
    } while ( rsize == -1 && errno == EINTR );
    
    if ( rsize > 0 ) {
      // position end one past buffer
      start = buffer;
      end = buffer + rsize;
    } else {
      // eof or error
      return ((-2)-rsize);
    }
  }

  return ((unsigned char)(*start));
}

#endif // USE_LIBBZ2
