//
// $Id: irr.cc,v 1.1 1999/08/05 21:06:01 voeckler Exp $
//
// Author:  Jens-S. Vckler <voeckler@rvs.uni-hannover.de>
//
// File:    irr.cc
//          Thu Aug 05 1999
//
// (c) 1999 Lehrgebiet Rechnernetze und Verteilte Systeme
//          Universitt Hannover, Germany
//
// Permission to use, copy, modify, distribute, and sell this software
// and its documentation for any purpose is hereby granted without fee,
// provided that (i) the above copyright notices and this permission
// notice appear in all copies of the software and related documentation,
// and (ii) the names of the Lehrgebiet Rechnernetze und Verteilte
// Systeme and the University of Hannover may not be used in any
// advertising or publicity relating to the software without the
// specific, prior written permission of Lehrgebiet Rechnernetze und
// Verteilte Systeme and the University of Hannover.
//
// THE SOFTWARE IS PROVIDED "AS-IS" AND WITHOUT WARRANTY OF ANY KIND,
// EXPRESS, IMPLIED OR OTHERWISE, INCLUDING WITHOUT LIMITATION, ANY
// WARRANTY OF MERCHANTABILITY OR FITNESS FOR A PARTICULAR PURPOSE.
//
// IN NO EVENT SHALL THE LEHRGEBIET RECHNERNETZE UND VERTEILTE SYSTEME OR
// THE UNIVERSITY OF HANNOVER BE LIABLE FOR ANY SPECIAL, INCIDENTAL,
// INDIRECT OR CONSEQUENTIAL DAMAGES OF ANY KIND, OR ANY DAMAGES
// WHATSOEVER RESULTING FROM LOSS OF USE, DATA OR PROFITS, WHETHER OR NOT
// ADVISED OF THE POSSIBILITY OF DAMAGE, AND ON ANY THEORY OF LIABILITY,
// ARISING OUT OF OR IN CONNECTION WITH THE USE OR PERFORMANCE OF THIS
// SOFTWARE.
//
// $Log: irr.cc,v $
// Revision 1.1  1999/08/05 21:06:01  voeckler
// Initial revision
//
//
#if defined(__GNUC__) || defined(__GNUG__)
#pragma implementation
#endif

#include <assert.h>
#include <sys/types.h>
#include <stdio.h>
#include <unistd.h>
#include <stdlib.h>
#include <time.h>
#include <fcntl.h>
#include <string.h>
#include <errno.h>

#include <sys/socket.h>
#include <netinet/in.h>
#include <arpa/inet.h>
#include <netdb.h>

#include "ctype.hh"
#include "irr.hh"

static const char* RCS_ID =
"$Id: irr.cc,v 1.1 1999/08/05 21:06:01 voeckler Exp $";

static
int
connectTo( MyUInt32 host )
{
  int sockfd = socket( PF_INET, SOCK_STREAM, IPPROTO_TCP );
  if ( sockfd == -1 ) {
    perror("socket");
    return -1;
  }

  struct sockaddr_in server;
  memset( &server, 0, sizeof(server) );
  server.sin_family = AF_INET;
  server.sin_addr.s_addr = host;
  server.sin_port = htons(43);
  if ( connect( sockfd, (struct sockaddr*) &server, sizeof(server) ) == -1 ) {
    fprintf( stderr, "connect(%s): %s\n", 
	     inet_ntoa(server.sin_addr), strerror(errno));
    close(sockfd);
    return -1;
  }

  return sockfd;
}

IRRCache::IRRCache( const char* cacheFile,
		    const char* whoisServer,
		    MyUInt32 posTTL,
		    MyUInt32 negTTL )
  :BaseCache(posTTL,negTTL),db(0),sockfd(-1)
{
  assert(cacheFile != 0);
  assert(whoisServer != 0);
  if ( (db = dbm_open( (char*) cacheFile, O_RDWR | O_CREAT, 0664 )) == 0 ) {
    fprintf( stderr, "dbm_open(%s): unable to open DB file!\n", cacheFile );
    exit(1);
  }

  if ( ISDIGIT(*whoisServer) ) {
    if ( (server = inet_addr(whoisServer)) == C_U32(-1) ) {
      fprintf( stderr, "\"%s\" is not a valid dotted quad!\n", whoisServer );
      exit(1);
    }
  } else {
    struct hostent* h = gethostbyname(whoisServer);
    if ( h ) {
      memcpy( &server, h->h_addr_list[0], sizeof(server) );
    } else {
      fprintf( stderr, "ghbn(%s): unable to resolve hostname\n", whoisServer );
      exit(1);
    }
  }

  // connect to whois server
  if ( (sockfd = connectTo(server)) == -1 ) exit(1);

  // establish persistent connection
  if ( write( sockfd, "!!\r\n", 4 ) < 4 ) {
    perror("write");
    exit(1);
  }
}

IRRCache::~IRRCache()
{
  if ( sockfd != -1 ) {
    write( sockfd, "!q\r\n", 4 );
    close(sockfd);
  }
  if ( dbm_error(db) ) dbm_clearerr(db);
  dbm_close(db);
}

inline
bool
check_junk( const char* s )
{
  return ( strncmp( s, "*rt: 0.0.0.0", 12 ) == 0 ||
	   strncmp( s, "*rt: 128.0.0.0", 14 ) == 0 ||
	   strncmp( s, "*rt: 192.0.0.0", 14 ) == 0 );
}

static
int
send_query_get_answer( int sockfd, const char* query,
		       char*& answer, size_t& size )
  // returns: offset into answer, negative flags error  
{
  // query server
  // FIXME: make error failsafe!
  int rsize, wsize = write( sockfd, query, strlen(query) );
    
  // obtain first char of answer
  char code;
  if ( read( sockfd, &code, 1 ) != 1 ) return 0;

  size = 0;
  if ( code == 'A' ) {
    // successful answer, read remainder of line --> size
    do {
      rsize = read( sockfd, &code, 1 );
      if ( rsize == -1 && errno == EINTR ) continue;
      if ( rsize == 1 && ISDIGIT(code) ) size = size*10 + (code-'0');
    } while ( rsize == 1 && code != '\n' );
      
    // read answer chunk
    answer = new char[size+1];
    MyUInt32 part = 0;
    do {
      rsize = read( sockfd, answer+part, size-part );
      if ( rsize == -1 && errno == EINTR ) continue;
      if ( rsize > 0 ) part += rsize;
    } while ( part < size && rsize > 0 );

    // read final character statement
    answer[size] = 0;
    read( sockfd, &code, 1 ); // should be 'C'
    read( sockfd, &code, 1 ); // should be '\n'

    // remove junk info
    // FIXME: junk is believed to start out the data
    char* start = answer;
    while ( check_junk(start) ) {
      start = strstr(start,"\n\n");
      if ( start == 0 ) start = answer+strlen(answer);
      else start += 2;
    }

    return start-answer;
  } else {
    // no success
    read( sockfd, &code, 1 ); // should be '\n'
    return -1;
  }
}


int
IRRCache::getOriginByHost( MyUInt32 address )
{
  // create class c network query
  struct in_addr ip;
  ip.s_addr = htonl(IN_CLASSC_NET & ntohl(address));

  // check with database
  MyUInt32 now = time(0);
  requests[C_ALL]++;
  MyUInt32 value[2] = { 0, 0 };
  datum _key;
  _key.dptr = (char*) &ip;
  _key.dsize = sizeof(ip);
  datum _value = dbm_fetch( db, _key );
  if ( _value.dptr ) {
    memcpy( value, _value.dptr, sizeof(value) );
#if USES_GDBM
    free((void*) _value.dptr );
#endif
  }

  if ( value[0] < now || _value.dptr == 0 ) {
    // assemble query
    char query[32];
    sprintf( query, "!r%s,l\r\n", inet_ntoa(ip) );
    
    // query server
    char* answer = 0;
    size_t size;
    int offset = send_query_get_answer( sockfd, query, answer, size );

    if ( offset >= 0 ) {
      // extract origin information
      int result = 0;
      for ( char* s = strstr( answer+offset, "\n*or:" ); 
	    s != 0; 
	    s = strstr( s+1, "\n*or:" ) ) {
	int temp(0);
	sscanf( s+2, "or: AS%d\n", &temp );
	if ( temp > 0 ) result = temp;
      }
      delete[] answer;
      
      // store into db file
      _key.dptr = (char*) &ip;
      _key.dsize = sizeof(ip);
      
      value[0] = now + positiveTTL;
      value[1] = result;
      _value.dptr = (char*) &value;
      _value.dsize = sizeof(value);
      dbm_store( db, _key, _value, DBM_REPLACE );
    } else {
      // no success --> enter failure into db
      _key.dptr = (char*) &ip;
      _key.dsize = sizeof(ip);
      value[0] = now + negativeTTL;
      value[1] = 0;
      _value.dptr = (char*) &value;
      _value.dsize = sizeof(value);
      dbm_store( db, _key, _value, DBM_REPLACE );
    }
  } else {
    requests[C_HIT]++;
  }

  // there was something in the db, use the db value
  return value[1];
}

int
IRRCache::getOriginByHost( DNSCache& dns, const char* hostname )
{
  if ( ISDIGIT(*hostname) ) {
    MyUInt32 temp = inet_addr(hostname);
    if ( temp == C_U32(-1) ) {
      requests[C_ALL]++; 
      return -1; 
    } else
      return getOriginByHost(temp);
  } else {
    char temp[256];
    if ( dns.gethostbyname(hostname,temp,sizeof(temp)) == 0 ) {
      requests[C_ALL]++;
      return -1;
    } else
      return getOriginByHost(inet_addr(temp));
  }
}

String 
IRRCache::getDescriptionByOrigin( MyUInt32 address )
{
  char temp[32];
  sprintf( temp, "AS" SF_U32, address );
  return getDescriptionByOrigin(temp);
}

String 
IRRCache::getDescriptionByOrigin( const char* address )
{
  // check with database
  MyUInt32 ttl = 0;
  MyUInt32 now = time(0);
  requests[C_ALL]++;
  String result;
  datum _key;

  _key.dptr = (char*) address;
  _key.dsize = strlen(address)+1;
  datum _value = dbm_fetch( db, _key );
  if ( _value.dptr ) {
    memcpy( &ttl, _value.dptr, sizeof(ttl) );
    result = String(((char*) _value.dptr)+sizeof(ttl) );
#if USES_GDBM
    free((void*) _value.dptr );
#endif
  }

  if ( ttl < now ) {
    // assemble query
    char query[32];
    sprintf( query, "!man,%s\r\n", address );
    
    // query server
    char* answer = 0;
    size_t size = 0;
    int offset = send_query_get_answer( sockfd, query, answer, size );
#if 0
    fprintf( stderr, "# lookup for %s %d %s\n", query, offset,
	     offset >= 0 ? answer : "(empty)" );
#endif
    if ( offset >= 0 ) {
      // extract description information
      char* p;
      for ( char* s = strstr( answer+offset, "\n*de:" ); 
	    s != 0; 
	    s = strstr( s+1, "\n*de:" ) ) {
	s += 5; // skip start
	if ( (p = strchr( s, '\n' )) ) 
	  result = String( s, p-s-1, 1 );
	else 
	  result = String( s );
	if ( (s = strstr( p+1, "\n\n" )) == 0 )
	  s = answer+size-1;
      }
      delete[] answer;
      ttl = now + positiveTTL;
    } else {
      // looup failure
      ttl = now + negativeTTL;
    }
    // store into db file
    _key.dptr = (char*) address;
    _key.dsize = strlen(address)+1;
    
    size = result.length() + 1 + sizeof(ttl);
    char* value = new char[size];
    memcpy( value, &ttl, sizeof(ttl) );
    memcpy( value+sizeof(ttl), result(), result.length()+1 );
    _value.dptr = value;
    _value.dsize = size;
    dbm_store( db, _key, _value, DBM_REPLACE );
    delete[] value;
  } else {
    requests[C_HIT]++;
  }

  // there was something in the db, use the db value
  return result;
}
