//
// $Id: offvector.hh,v 1.1 1999/08/05 21:17:11 voeckler Exp $
//
// Author:  Jens-S. Vckler <voeckler@rvs.uni-hannover.de>
//
// File:    offvector.hh
//          Thu Aug 05 1999
//
// (c) 1999 Lehrgebiet Rechnernetze und Verteilte Systeme
//          Universitt Hannover, Germany
//
// Permission to use, copy, modify, distribute, and sell this software
// and its documentation for any purpose is hereby granted without fee,
// provided that (i) the above copyright notices and this permission
// notice appear in all copies of the software and related documentation,
// and (ii) the names of the Lehrgebiet Rechnernetze und Verteilte
// Systeme and the University of Hannover may not be used in any
// advertising or publicity relating to the software without the
// specific, prior written permission of Lehrgebiet Rechnernetze und
// Verteilte Systeme and the University of Hannover.
//
// THE SOFTWARE IS PROVIDED "AS-IS" AND WITHOUT WARRANTY OF ANY KIND,
// EXPRESS, IMPLIED OR OTHERWISE, INCLUDING WITHOUT LIMITATION, ANY
// WARRANTY OF MERCHANTABILITY OR FITNESS FOR A PARTICULAR PURPOSE.
//
// IN NO EVENT SHALL THE LEHRGEBIET RECHNERNETZE UND VERTEILTE SYSTEME OR
// THE UNIVERSITY OF HANNOVER BE LIABLE FOR ANY SPECIAL, INCIDENTAL,
// INDIRECT OR CONSEQUENTIAL DAMAGES OF ANY KIND, OR ANY DAMAGES
// WHATSOEVER RESULTING FROM LOSS OF USE, DATA OR PROFITS, WHETHER OR NOT
// ADVISED OF THE POSSIBILITY OF DAMAGE, AND ON ANY THEORY OF LIABILITY,
// ARISING OUT OF OR IN CONNECTION WITH THE USE OR PERFORMANCE OF THIS
// SOFTWARE.
//
// $Log: offvector.hh,v $
// Revision 1.1  1999/08/05 21:17:11  voeckler
// Initial revision
//
//
#ifndef _OFFVECTOR_HH
#define _OFFVECTOR_HH

#if defined(__GNUC__) || defined(__GNUG__)
#pragma interface
#endif

#include "typedefs.h"
#include "counter.hh"

template <class T> class OffsetVector {
public:
  inline OffsetVector()
    :array(0),_size(0),offset(0) { }
  inline ~OffsetVector()
    { delete[] array; array = 0; }
  inline OffsetVector( const OffsetVector& gv )
    :array( new T[gv._size] ),_size(gv._size),offset(gv.offset)
    { for ( int i=0; i<_size; i++ ) array[i] = gv.array[i]; }

  inline size_t size() const { return _size; }
  inline MySInt32 getOffset() const { return offset; }
  
  OffsetVector& operator=( const OffsetVector& gv )
    {
      if ( &gv != this ) {
	delete[] array;
	_size = gv._size;
	offset = gv.offset;
	for ( int i=0; i<_size; i++ ) 
	  array[i] = gv.array[i];
      }
      return *this;
    }

  T& operator[]( MySInt32 index )
    {
      MySInt32 realindex = index - offset;
      if ( array == 0 ) {
	// first growth
	offset = index;
	array = new T[ (_size=32) ];
	realindex = 0;
      } else {
	if ( realindex < 0 ) {
	  // negative growth necessary
	  size_t newsize = _size - realindex;
	  MySInt32 newoffset = index;
	  T* newarray = new T[newsize];
	  for ( size_t i=0; i<_size; ++i ) newarray[i-realindex] = array[i];
	  delete[] array;
	  _size = newsize;
	  offset = newoffset;
	  array = newarray;
	  realindex = 0;
	} else if ( realindex >= _size ) {
	  // positive growth necessary
	  size_t newsize = _size << 1;
	  while ( realindex >= newsize ) newsize <<= 1;
	  T* newarray = new T[newsize];
	  for ( size_t i=0; i<_size; ++i ) newarray[i] = array[i];
	  delete[] array;
	  _size = newsize;
	  array = newarray;
	}
      }
      return array[realindex];
    }

  inline T operator[]( MySInt32 index ) const
    {
      MySInt32 realindex = index - offset;
      if ( array == 0 || realindex < 0 || realindex >= _size ) 
	return T();
      else
	return array[realindex];
    }

protected:
  T*        array;
  size_t    _size;
  MySInt32  offset;
};

#endif // _OFFVECTOR_HH
