//
// $Id: output.cc,v 1.1 1999/08/05 21:17:11 voeckler Exp $
//
// Author:  Jens-S. Vckler <voeckler@rvs.uni-hannover.de>
//
// File:    output.cc
//          Thu Aug 05 1999
//
// (c) 1999 Lehrgebiet Rechnernetze und Verteilte Systeme
//          Universitt Hannover, Germany
//
// Permission to use, copy, modify, distribute, and sell this software
// and its documentation for any purpose is hereby granted without fee,
// provided that (i) the above copyright notices and this permission
// notice appear in all copies of the software and related documentation,
// and (ii) the names of the Lehrgebiet Rechnernetze und Verteilte
// Systeme and the University of Hannover may not be used in any
// advertising or publicity relating to the software without the
// specific, prior written permission of Lehrgebiet Rechnernetze und
// Verteilte Systeme and the University of Hannover.
//
// THE SOFTWARE IS PROVIDED "AS-IS" AND WITHOUT WARRANTY OF ANY KIND,
// EXPRESS, IMPLIED OR OTHERWISE, INCLUDING WITHOUT LIMITATION, ANY
// WARRANTY OF MERCHANTABILITY OR FITNESS FOR A PARTICULAR PURPOSE.
//
// IN NO EVENT SHALL THE LEHRGEBIET RECHNERNETZE UND VERTEILTE SYSTEME OR
// THE UNIVERSITY OF HANNOVER BE LIABLE FOR ANY SPECIAL, INCIDENTAL,
// INDIRECT OR CONSEQUENTIAL DAMAGES OF ANY KIND, OR ANY DAMAGES
// WHATSOEVER RESULTING FROM LOSS OF USE, DATA OR PROFITS, WHETHER OR NOT
// ADVISED OF THE POSSIBILITY OF DAMAGE, AND ON ANY THEORY OF LIABILITY,
// ARISING OUT OF OR IN CONNECTION WITH THE USE OR PERFORMANCE OF THIS
// SOFTWARE.
//
// $Log: output.cc,v $
// Revision 1.1  1999/08/05 21:17:11  voeckler
// Initial revision
//
//

#if defined(__GNUC__) || defined(__GNUG__)
#pragma implementation
#endif

#include <time.h>
#include <string.h>
#include <stdlib.h>
#include "ctype.hh"
#include "string.hh"
#include "output.hh"
#include "tools.hh"
#include "global.hh"

static const char* RCS_ID =
"$Id: output.cc,v 1.1 1999/08/05 21:17:11 voeckler Exp $";

static String str_other( "<%d more>" );
static const char* timeFormat = "%d.%m.%Y %H:%M:%S";

void 
showTitle()
{
  char s[80];

  time_t sometime = globals.startTime;
  strftime( s, sizeof(s), "%c %Z", localtime( &sometime ) );
  printf( "# From \"%s\" ", s );
  sometime = globals.stopTime;
  strftime( s, sizeof(s), "%c %Z", localtime( &sometime ) );
  printf( "till \"%s\"\n\n", s );
}

void
showOverview( const Counters& counters )
{
  char s[64];
  Counter temp;

  puts( valueHead( "# SUM", 6, true )() );
  puts( valueTail( 6, true )() );
  printf( "UDP    %s\n", valueBody(counters.udp, counters.all, counters.udp_hit)() );
  temp += counters.udp_hit;
  printf( "TCP    %s\n", valueBody(counters.tcp, counters.all, counters.tcp_hit)() );
  temp += counters.tcp_hit;
  puts( valueTail( 6, true )() );
  printf( "ALL    %s\n", valueBody(counters.all, counters.all, temp)() );
  puts("");
  
  puts( valueHead( "# UDP BY STATUS", 20, false )() );
  puts( valueTail( 20, false )() );
  printf( "%-20s %s\n", "HIT", valueBody(counters.udp_hit,counters.udp)() );
  size_t size = LogLine::stable.distinct()+1; // include zeroth element!
  size_t* index = indexByReqs( counters.udp_hit_status, size );
  for ( MyUInt32 i=0; i<size; ++i )
    if ( counters.udp_hit_status[index[i]].reqs() )
      printf( " %-19s %s\n", 
	      LogLine::stable.reverse(s,sizeof(s),index[i]),
	      valueBody(counters.udp_hit_status[index[i]],counters.udp)() );
  delete[] index;
  printf( "%-20s %s\n", "MISS", valueBody(counters.udp_miss,counters.udp)() );
  index = indexByReqs( counters.udp_miss_status, size );
  for ( MyUInt32 i=0; i<size; ++i )
    if ( counters.udp_miss_status[index[i]].reqs() )
      printf( " %-19s %s\n", 
	      LogLine::stable.reverse(s,sizeof(s),index[i]),
	      valueBody(counters.udp_miss_status[index[i]],counters.udp)() );
  delete[] index;
  puts( valueTail( 20, false )() );
  printf( "%-20s %s\n", "SUM", valueBody(counters.udp,counters.udp)() );
  puts("");



  puts( valueHead( "# TCP BY STATUS", 20, false )() );
  puts( valueTail( 20, false )() );
  printf( "%-20s %s\n", "HIT", valueBody(counters.tcp_hit,counters.tcp)() );
  index = indexByReqs( counters.tcp_hit_status, size );
  for ( MyUInt32 i=0; i<size; ++i )
    if ( counters.tcp_hit_status[index[i]].reqs() )
      printf( " %-19s %s\n", 
	      LogLine::stable.reverse(s,sizeof(s),index[i]),
	      valueBody(counters.tcp_hit_status[index[i]],counters.tcp)() );
  delete[] index;
  printf( "%-20s %s\n", "MISS", valueBody(counters.tcp_miss,counters.tcp)() );
  index = indexByReqs( counters.tcp_miss_status, size );
  for ( MyUInt32 i=0; i<size; ++i )
    if ( counters.tcp_miss_status[index[i]].reqs() )
      printf( " %-19s %s\n", 
	      LogLine::stable.reverse(s,sizeof(s),index[i]),
	      valueBody(counters.tcp_miss_status[index[i]],counters.tcp)() );
  delete[] index;
  printf( "%-20s %s\n", "NONE", valueBody(counters.tcp_miss_none,counters.tcp)() );
  index = indexByReqs( counters.tcp_miss_none_status, size );
  for ( MyUInt32 i=0; i<size; ++i )
    if ( counters.tcp_miss_none_status[index[i]].reqs() )
      printf( " %-19s %s\n", 
	      LogLine::stable.reverse(s,sizeof(s),index[i]),
	      valueBody(counters.tcp_miss_none_status[index[i]],counters.tcp)() );
  delete[] index;
  puts( valueTail( 20, false )() );
  printf( "%-20s %s\n", "SUM", valueBody(counters.tcp,counters.tcp)() );
  puts("");
}

void
showMethods( const Counters& counters )
{
  char s[64];
  Counter thit, temp;
  size_t size = LogLine::mtable.distinct()+1; // include zeroth element!
  size_t* index = indexByReqs( counters.methods, size );

  puts( valueHead("# METHODS", 10, true)() );
  puts( valueTail(10,true)() );
  for ( MyUInt32 i=0; i<size; ++i ) {
    if ( counters.methods[index[i]].reqs() ) {
      printf( "%-10s %s\n", LogLine::mtable.reverse(s,sizeof(s),index[i]), 
	      valueBody( counters.methods[index[i]],
			 counters.all,
			 counters.hit_methods[index[i]] )() );
      temp += counters.methods[index[i]];
      thit += counters.hit_methods[index[i]];
    }
  }
  delete[] index;
  puts( valueTail(10,true)() );
  printf( "%-10s %s\n\n", "SUM", valueBody(temp,counters.all,thit)() );
  if ( temp != counters.all || thit != counters.tcp_hit+counters.udp_hit )
    printf( "%-10s %s\n\n", "SHOULD-BE", 
	    valueBody(counters.all,counters.all,
		      counters.tcp_hit+counters.udp_hit)() );
}

void showSchemes( const Counters& counters )
{
  char s[64];
  Counter thit, temp;
  size_t size = LogLine::schemes.distinct()+1; // include zeroth element!
  size_t* index = indexByReqs( counters.tcp_scheme, size );

  puts( valueHead("# SCHEMES", 16, true)() );
  puts( valueTail(16,true)() );
  for ( MyUInt32 i=0; i<size; ++i ) {
    if ( counters.tcp_scheme[index[i]].reqs() ) {
      printf( "%-16s %s\n", LogLine::schemes.reverse(s,sizeof(s),index[i]), 
	      valueBody( counters.tcp_scheme[index[i]],
			 counters.tcp,
			 counters.tcp_hit_scheme[index[i]] )() );
      temp += counters.tcp_scheme[index[i]];
      thit += counters.tcp_hit_scheme[index[i]];
    }
  }
  delete[] index;
  puts( valueTail(16,true)() );
  printf( "%-16s %s\n\n", "SUM", valueBody(temp,counters.tcp,thit)() );
  if ( temp != counters.tcp || thit != counters.tcp_hit )
    printf( "%-16s %s\n\n", "SHOULD-BE", 
	    valueBody(counters.tcp,counters.tcp,
		      counters.tcp_hit)() );
}

void
showTLD( const Counters& counters, const size_t number )
{
  Counter thit;
  Counter temp;
  Counter leftover;
  Counter lhitover;
  MyUInt32 n = C_U32(0);

  puts( valueHead( "# TLD BY REQ.", 13, true )() );
  puts( valueTail( 13, true )() );
  size_t* index = indexByReqs( counters.tcp_tld );
  for ( size_t i=0; i<counters.tcp_tld.size(); ++i ) {
    String key( counters.tcp_tld.key(index[i]) );
    if ( number == 0 || number > i ) {
      printf( "%-13s %s\n", key(),
	      valueBody( counters.tcp_tld[index[i]], 
			 counters.tcp,
			 counters.tcp_hit_tld[key] )() );
    } else {
      leftover += counters.tcp_tld[index[i]];
      lhitover += counters.tcp_hit_tld[key];
      n++;
    }
    temp += counters.tcp_tld[index[i]];
    thit += counters.tcp_hit_tld[key];
  }
  delete[] index;

  if ( n ) {
    char here[16];
    sprintf( here, str_other(), n );
    printf( "%-13s %s\n", here, 
	    valueBody( leftover, counters.tcp, lhitover )() );
  }
  puts( valueTail( 13, true )() );
  printf( "%-13s %s\n", "SUM", valueBody(temp,counters.tcp,thit)() );
  if ( temp != counters.tcp || thit != counters.tcp_hit )
    printf( "%-13s %s\n", "SHOULD-BE", 
	    valueBody(counters.tcp,counters.tcp,counters.tcp_hit)() );
  puts("");




  puts( valueHead( "# TLD BY VOL.", 13, true )() );
  puts( valueTail( 13, true )() );
  index = indexBySize( counters.tcp_tld );
  thit = temp = Counter();
  leftover = lhitover = Counter();
  n = C_U32(0);
  for ( size_t i=0; i<counters.tcp_tld.size(); ++i ) {
    String key( counters.tcp_tld.key(index[i]) );
    if ( number == 0 || number > i ) {
      printf( "%-13s %s\n", key(),
	      valueBody( counters.tcp_tld[index[i]], 
			 counters.tcp,
			 counters.tcp_hit_tld[key] )() );
    } else {
      leftover += counters.tcp_tld[index[i]];
      lhitover += counters.tcp_hit_tld[key];
      n++;
    }
    temp += counters.tcp_tld[index[i]];
    thit += counters.tcp_hit_tld[key];
  }
  delete[] index;

  if ( n ) {
    char here[16];
    sprintf( here, str_other(), n );
    printf( "%-13s %s\n", here, 
	    valueBody( leftover, counters.tcp, lhitover )() );
  }
  puts( valueTail( 13, true )() );
  printf( "%-13s %s\n", "SUM", valueBody(temp,counters.tcp,thit)() );
  if ( temp != counters.tcp || thit != counters.tcp_hit )
    printf( "%-13s %s\n", "SHOULD-BE", 
	    valueBody(counters.tcp,counters.tcp,counters.tcp_hit)() );
  puts("");
}

void
showSLD( const Counters& counters, const size_t number )
{
  Counter thit;
  Counter temp;
  Counter leftover;
  Counter lhitover;
  MyUInt32 n = C_U32(0);

  puts( valueHead( "# 2ND LEVEL DOMAIN BY REQEST", 33, true )() );
  puts( valueTail( 33, true )() );
  size_t* index = indexByReqs( counters.tcp_sld );
  for ( size_t i=0; i<counters.tcp_sld.size(); ++i ) {
    String key( counters.tcp_sld.key(index[i]) );
    if ( number == 0 || number > i ) {
      printf( "%-33s %s\n", key(),
	      valueBody( counters.tcp_sld[index[i]], 
			 counters.tcp,
			 counters.tcp_hit_sld[key] )() );
    } else {
      leftover += counters.tcp_sld[index[i]];
      lhitover += counters.tcp_hit_sld[key];
      n++;
    }
    temp += counters.tcp_sld[index[i]];
    thit += counters.tcp_hit_sld[key];
  }
  delete[] index;

  if ( n ) {
    char here[16];
    sprintf( here, str_other(), n );
    printf( "%-33s %s\n", here, 
	    valueBody( leftover, counters.tcp, lhitover )() );
  }
  puts( valueTail( 33, true )() );
  printf( "%-33s %s\n", "SUM", valueBody(temp,counters.tcp,thit)() );
  if ( temp != counters.tcp || thit != counters.tcp_hit )
    printf( "%-33s %s\n", "SHOULD-BE", 
	    valueBody(counters.tcp,counters.tcp,counters.tcp_hit)() );
  puts("");




  puts( valueHead( "# 2ND LEVEL DOMAIN BY VOLUME", 33, true )() );
  puts( valueTail( 33, true )() );
  index = indexBySize( counters.tcp_sld );
  thit = temp = Counter();
  leftover = lhitover = Counter();
  n = C_U32(0);
  for ( size_t i=0; i<counters.tcp_sld.size(); ++i ) {
    String key( counters.tcp_sld.key(index[i]) );
    if ( number == 0 || number > i ) {
      printf( "%-33s %s\n", key(),
	      valueBody( counters.tcp_sld[index[i]], 
			 counters.tcp,
			 counters.tcp_hit_sld[key] )() );
    } else {
      leftover += counters.tcp_sld[index[i]];
      lhitover += counters.tcp_hit_sld[key];
      n++;
    }
    temp += counters.tcp_sld[index[i]];
    thit += counters.tcp_hit_sld[key];
  }
  delete[] index;

  if ( n ) {
    char here[16];
    sprintf( here, str_other(), n );
    printf( "%-33s %s\n", here, 
	    valueBody( leftover, counters.tcp, lhitover )() );
  }
  puts( valueTail( 33, true )() );
  printf( "%-33s %s\n", "SUM", valueBody(temp,counters.tcp,thit)() );
  if ( temp != counters.tcp || thit != counters.tcp_hit )
    printf( "%-33s %s\n", "SHOULD-BE", 
	    valueBody(counters.tcp,counters.tcp,counters.tcp_hit)() );
  puts("");
}

extern "C" {
static
int 
myCompare( const void* a, const void* b )
{ 
  size_t aa = *((size_t*) a);
  size_t bb = *((size_t*) b);
  fprintf( stderr, "<< %d <=> %d >>\n", aa, bb );
  return aa - bb;
}
}

static
String
maybeResolve( DNSCache& dns, const String& what )
{
  if ( ISDIGIT(what[0]) && globals.clientFQDN ) {
    char temp[256];
    if ( dns.gethostbyaddr( what(), temp, sizeof(temp) ) )
      return String(temp);
  }
  return String(what);
}

void
showHierarchy( DNSCache& dns, const Counters& counters )
{
  char s[80];
  Counter temp, thit;
  size_t size = LogLine::htable.distinct()+1; // include zeroth element!

  puts( valueHead( "# HIERARCHY OVERVIEW", 32, true )() );
  puts( valueTail( 32, true )() );
  printf( "%-32s %s\n", "DIRECT to source", 
	  valueBody( counters.hier_direct, counters.hier )() );
  for ( MyUInt32 i=0; i<size; ++i ) 
    if ( counters.hier_direct_method[i].reqs() ) {
      printf( " %-31s %s\n", LogLine::htable.reverse(s,sizeof(s),i),
	      valueBody( counters.hier_direct_method[i],
			 counters.hier,
			 counters.tcp_hit_hier[i] )() );
      temp += counters.hier_direct_method[i];
      thit += counters.tcp_hit_hier[i];
    }

  printf( "%-32s %s\n", "PARENT fetches", 
	  valueBody( counters.hier_parent, counters.hier )() );
  for ( MyUInt32 i=0; i<size; ++i ) 
    if ( counters.hier_parent_method[i].reqs() ) {
      printf( " %-31s %s\n", LogLine::htable.reverse(s,sizeof(s),i),
	      valueBody( counters.hier_parent_method[i],
			 counters.hier,
			 counters.tcp_hit_hier[i] )() );
      temp += counters.hier_parent_method[i];
      thit += counters.tcp_hit_hier[i];
    }

  printf( "%-32s %s\n", "PEER fetches", 
	  valueBody( counters.hier_peer, counters.hier )() );
  for ( MyUInt32 i=0; i<size; ++i ) 
    if ( counters.hier_peer_method[i].reqs() ) {
      printf( " %-31s %s\n", LogLine::htable.reverse(s,sizeof(s),i),
	      valueBody( counters.hier_peer_method[i],
			 counters.hier,
			 counters.tcp_hit_hier[i] )() );
      temp += counters.hier_peer_method[i];
      thit += counters.tcp_hit_hier[i];
    }

  puts( valueTail( 32, true )() );
  printf( "%-32s %s\n", "SUM", valueBody(temp,counters.hier,thit)() );
  if ( temp != counters.hier )
    printf( "%-32s %s\n", "SHOULD-BE", 
	    valueBody(counters.hier,counters.hier,counters.tcp_hit)() );
  puts("");

  // -----------------------------------------------------------------

  puts( valueHead( "# HIERARCHY BY HOST", 40, false )() );
  puts( valueTail( 40, false )() );
  printf( "%-40s %s\n", "DIRECT to source", 
	  valueBody( counters.hier_direct, counters.hier )() );

  size_t* index = indexByReqs( counters.hier_peer_host );
  Counter* sub = new Counter[size];
  temp = counters.hier_direct;
  for ( size_t i=0; i<counters.hier_peer_host.size(); ++i ) {
    String key( counters.hier_peer_host.key(index[i]) );
    printf( "%-40s %s\n", maybeResolve(dns,key)(),
	    valueBody( counters.hier_peer_host[index[i]], 
		       counters.hier )() );
    temp += counters.hier_peer_host[index[i]];
    
    for ( size_t j=0; j<size; ++j ) 
      sub[j] = counters.hier_peer_status[j][key];
    size_t* local = indexByReqs(sub,size);
    for ( size_t j=0; j<size; ++j ) {
      if ( sub[local[j]].reqs() ) 
	printf( " %-39s %s\n", LogLine::htable.reverse(s,sizeof(s),local[j]),
		valueBody( sub[local[j]], counters.hier )() );
    }
    delete[] local;
  }
  delete[] sub;
  delete[] index;
  puts( valueTail( 40, false )() );
  printf( "%-40s %s\n", "SUM", valueBody(temp,counters.hier)() );
  if ( temp != counters.hier )
    printf( "%-40s %s\n", "SHOULD-BE", 
	    valueBody(counters.hier,counters.hier)() );
  puts("");
}




void
showMediatypes( const Counters& counters )
{
  char s[80];
  Counter thit;
  Counter temp;

  puts( valueHead( "# MEDIATYPES BY REQUESTS", 24, true )() );
  puts( valueTail( 24, true )() );
  size_t* index = indexByReqs( counters.tcp_mime, counters.nMedia );
  for ( size_t i=0; i<counters.nMedia; ++i ) {
    if ( counters.tcp_mime[index[i]].reqs() ) {
      printf( "%-24s %s\n", 
	      LogLine::mediatypes.reverse(s,sizeof(s),index[i]),
	      valueBody( counters.tcp_mime[index[i]], 
			 counters.tcp,
			 counters.tcp_hit_mime[index[i]] )() );
      if ( counters.tcp_submime[index[i]] ) {
	size_t* sub = indexByReqs( counters.tcp_submime[index[i]],
				   counters.nSubtype[index[i]] );
	for ( size_t j=0; j<counters.nSubtype[index[i]]; ++j ) {
	  if ( counters.tcp_submime[index[i]][sub[j]].reqs() ) {
	    printf( " /%-22s %s\n",
		    LogLine::subtypes[index[i]].reverse(s,sizeof(s),sub[j]),
		    valueBody( counters.tcp_submime[index[i]][sub[j]],
			       counters.tcp,
			       counters.tcp_hit_submime[index[i]][sub[j]] )() );
	  }
	}
	delete[] sub;
      }
    }
    temp += counters.tcp_mime[index[i]];
    thit += counters.tcp_hit_mime[index[i]];
  }
  delete[] index;
  puts( valueTail( 24, true )() );
  printf( "%-24s %s\n", "SUM", valueBody(temp,counters.tcp,thit)() );
  if ( temp != counters.tcp || thit != counters.tcp_hit )
    printf( "%-24s %s\n", "SHOULD-BE", 
	    valueBody(counters.tcp,counters.tcp,counters.tcp_hit)() );
  puts("");




  puts( valueHead( "# MEDIATYPES BY VOLUME", 24, true )() );
  puts( valueTail( 24, true )() );
  index = indexBySize( counters.tcp_mime, counters.nMedia );
  thit = temp = Counter();
  for ( size_t i=0; i<counters.nMedia; ++i ) {
    if ( counters.tcp_mime[index[i]].reqs() ) {
      printf( "%-24s %s\n", 
	      LogLine::mediatypes.reverse(s,sizeof(s),index[i]),
	      valueBody( counters.tcp_mime[index[i]], 
			 counters.tcp,
			 counters.tcp_hit_mime[index[i]] )() );
      if ( counters.tcp_submime[index[i]] ) {
	size_t* sub = indexBySize( counters.tcp_submime[index[i]],
				   counters.nSubtype[index[i]] );
	for ( size_t j=0; j<counters.nSubtype[index[i]]; ++j ) {
	  if ( counters.tcp_submime[index[i]][sub[j]].reqs() ) {
	    printf( " /%-22s %s\n",
		    LogLine::subtypes[index[i]].reverse(s,sizeof(s),sub[j]),
		    valueBody( counters.tcp_submime[index[i]][sub[j]],
			       counters.tcp,
			       counters.tcp_hit_submime[index[i]][sub[j]] )() );
	  }
	}
	delete[] sub;
      }
    }
    temp += counters.tcp_mime[index[i]];
    thit += counters.tcp_hit_mime[index[i]];
  }
  delete[] index;
  puts( valueTail( 24, true )() );
  printf( "%-24s %s\n", "SUM", valueBody(temp,counters.tcp,thit)() );
  if ( temp != counters.tcp || thit != counters.tcp_hit )
    printf( "%-24s %s\n", "SHOULD-BE", 
	    valueBody(counters.tcp,counters.tcp,counters.tcp_hit)() );
  puts("");
}



void
showPorts( const Counters& counters )
{
  char temp[80];
  Counter sum;
  // with without port 80, which is stored in ports[64]...
  for ( int i=0; i<64; ++i ) sum += counters.tcp_ports[i];
  puts( valueHead( "# NON-STD PORTS", 15, false )() );
  puts( valueTail( 15, false )() );
  for ( int i=0; i<64; ++i )
    if ( counters.tcp_ports[i].reqs() ) {
      sprintf( temp, "%5d-%-5d", i << 10, ((i+1)<<10)-1 );
      if ( i==0 ) strcat( temp, "\\80" );
      printf( "%-15s %s\n", temp, 
	      valueBody( counters.tcp_ports[i], sum )() );
    }
  puts( valueTail( 15, false )() );
  Counter total = sum + counters.tcp_ports[64];
  printf( "%-15s %s\n", "NON-STD SUM", 
	  valueBody( sum, total )() );
  printf( "%-15s %s\n", "Port 80", 
	  valueBody( counters.tcp_ports[64], total )() );
  puts( valueTail( 15, false )() );
  printf( "%-15s %s\n", "SUM", valueBody( total, total )() );
  puts("");
}



void 
showUDPClients( DNSCache& dns, const Counters& counters, const size_t number )
{
  Counter thit;
  Counter temp;
  Counter leftover;
  Counter lhitover;
  MyUInt32 n = C_U32(0);

  puts( valueHead( "# UDP CLIENTS BY REQUESTS", 40, true )() );
  puts( valueTail( 40, true )() );
  size_t size = counters.udp_client.size();
  size_t* index = indexByReqs( counters.udp_client );
  for ( size_t i=0; i<size; ++i ) {
    if ( counters.udp_client[index[i]].reqs() ) {
      String key(counters.udp_client.key(index[i]));
      if ( number == 0 || number > i ) {
	printf( "%-40s %s\n", maybeResolve(dns,key).substring(0,40)(),
		valueBody( counters.udp_client[index[i]], 
			   counters.udp,
			   counters.udp_hit_client[key] )() );
      } else {
	leftover += counters.udp_client[index[i]];
	lhitover += counters.udp_hit_client[key];
	n++;
      }
      temp += counters.udp_client[index[i]];
      thit += counters.udp_hit_client[key];
    }
  }
  delete[] index;

  if ( n ) {
    char here[80];
    sprintf( here, str_other(), n );
    printf( "%-40s %s\n", here,
	    valueBody( leftover, counters.udp, lhitover )() );
  }
  puts( valueTail( 40, true )() );
  printf( "%-40s %s\n", "SUM", valueBody(temp,counters.udp,thit)() );
  if ( temp != counters.udp || thit != counters.udp_hit )
    printf( "%-40s %s\n", "SHOULD-BE", 
	    valueBody(counters.udp,counters.udp,counters.udp_hit)() );
  puts("");
}


void 
showTCPClients( DNSCache& dns, const Counters& counters, const size_t number )
{
  Counter  temp,  leftover;
  Counter  thit,  lhitover;
  Counter  terr,  lerrover;
  Counter tmiss, lmissover;
  MyUInt32 n = C_U32(0);

  puts( valueHead( "# TCP CLIENTS BY REQUST", 40, true ).substitute("HIT","HME",true)() );
  puts( valueTail( 40, true )() );

  size_t size = counters.tcp_client.size();
  size_t* index = indexByReqs( counters.tcp_client );
  for ( size_t i=0; i<size; ++i ) {
    if ( counters.tcp_client[index[i]].reqs() ) {
      String key(counters.tcp_client.key(index[i]));
      if ( number == 0 || number > i ) {
	printf( "%-40s %s\n", maybeResolve(dns,key).substring(0,40)(),
		valueBody( counters.tcp_client[index[i]], 
			   counters.tcp,
			   counters.tcp_hit_client[key] )() );
	if ( counters.tcp_miss_client[key].reqs() )
	  printf( " %-83s %s\n", "MISS",
		  valueBody( counters.tcp_miss_client[key],
			     counters.tcp_client[index[i]] )() );
	if ( counters.tcp_miss_none_client[key].reqs() )
	  printf( " %-83s %s\n", "NONE",
		  valueBody( counters.tcp_miss_none_client[key],
			     counters.tcp_client[index[i]] )() );
      } else {
	leftover += counters.tcp_client[index[i]];
	lhitover += counters.tcp_hit_client[key];
	lmissover += counters.tcp_miss_client[key];
	lerrover += counters.tcp_miss_none_client[key];
	n++;
      }
      temp += counters.tcp_client[index[i]];
      thit += counters.tcp_hit_client[key];
      tmiss += counters.tcp_miss_client[key];
      terr += counters.tcp_miss_none_client[key];
    }
  }
  delete[] index;

  if ( n ) {
    char here[80];
    sprintf( here, str_other(), n );
    printf( "%-40s %s\n", here, valueBody(leftover,counters.tcp,lhitover)() );
    if ( lmissover.reqs() )
      printf( " %-83s %s\n", "MISS", valueBody(lmissover,counters.tcp)() );
    if ( lerrover.reqs() )
      printf( " %-83s %s\n", "NONE", valueBody(lerrover,counters.tcp)() );
  }

  puts( valueTail( 40, true )() );
  printf( "%-40s %s\n", "SUM", valueBody(temp,counters.tcp,thit)() );
  if ( tmiss.reqs() )
    printf( " %-83s %s\n", "MISS-SUM", valueBody(tmiss,counters.tcp)() );
  if ( terr.reqs() )
    printf( " %-83s %s\n", "NONE-SUM", valueBody(terr,counters.tcp)() );
  if ( temp != counters.tcp || thit != counters.tcp_hit ||
       terr != counters.tcp_miss_none || tmiss != counters.tcp_miss ) {
    printf( "%-40s %s\n", "SHOULD-BE", 
	    valueBody(counters.tcp,counters.tcp,counters.tcp_hit)() );
    if ( counters.tcp_miss.reqs() )
      printf( " %-83s %s\n", "SHOULD-MISS", 
	      valueBody(counters.tcp_miss,counters.tcp)() );
    if ( counters.tcp_miss_none.reqs() )
      printf( " %-83s %s\n", "SHOULD-NONE", 
	      valueBody(counters.tcp_miss_none,counters.tcp)() );
  }
  puts("");
}

void
showASN( DNSCache& dns, IRRCache& irr, 
	 const Counters& counters, const size_t number )
{
  Counter temp;
  Counter leftover;
  MyUInt32 n = C_U32(0);
  CountMap asnmap;
  String host;

  // create ASN map
  for ( CountMapIterator i(counters.hier_direct_host); i.avail(); i++ ) {
    if ( i().reqs() ) {
      char temp[32];
      int asn = irr.getOriginByHost( dns, i.key()() );
      if ( asn < 0 ) {
	// DNS failure
	strcpy( temp, "<NODNS>" );
      } else if ( asn == 0 ) {
	// IRR failure
	strcpy( temp, "<NOIRR>" );
      } else {
	// successful lookup
	sprintf( temp, "AS%d", asn );
      }
      asnmap[String(temp)] += i();
    }
  }

  puts( valueHead( "# DIRECT DST BY REQUESTS", 40, false )() );
  puts( valueTail( 40, false )() );
  size_t size = asnmap.size();
  size_t* index = indexByReqs( asnmap );
  for ( size_t i=0; i<size; ++i ) {
    String key(asnmap.key(index[i]));
    if ( number == 0 || number > i ) {
      printf( "%-7s \"%-30s\" %s\n", key(),
	      irr.getDescriptionByOrigin(key()).substring(0,30)(),
	      valueBody( asnmap[size_t(index[i])], counters.hier )() );
    } else {
      leftover += asnmap[size_t(index[i])];
      n++;
    }
    temp += asnmap[index[i]];
  }
  delete[] index;

  if ( n ) {
    char here[80];
    sprintf( here, str_other(), n );
    printf( "%-40s %s\n", here, valueBody( leftover, counters.hier_direct )() );
  }
  puts( valueTail( 40, false )() );
  printf( "%-40s %s\n", "SUM", valueBody(temp,counters.hier_direct)() );
  if ( temp != counters.hier_direct )
    printf( "%-40s %s\n", "SHOULD-BE", 
	    valueBody(counters.hier_direct,counters.hier_direct)() );
  puts("");



  temp = leftover = Counter();
  n = C_U32(0);
  puts( valueHead( "# DIRECT DST BY VOLUME", 40, false )() );
  puts( valueTail( 40, false )() );
  index = indexBySize( asnmap );
  for ( size_t i=0; i<size; ++i ) {
    String key(asnmap.key(index[i]));
    if ( number == 0 || number > i ) {
      printf( "%-7s \"%-30s\" %s\n", key(),
	      irr.getDescriptionByOrigin(key()).substring(0,30)(),
	      valueBody( asnmap[size_t(index[i])], counters.hier )() );
    } else {
      leftover += asnmap[size_t(index[i])];
      n++;
    }
    temp += asnmap[index[i]];
  }
  delete[] index;

  if ( n ) {
    char here[80];
    sprintf( here, str_other(), n );
    printf( "%-40s %s\n", here, valueBody( leftover, counters.hier_direct )() );
  }
  puts( valueTail( 40, false )() );
  printf( "%-40s %s\n", "SUM", valueBody(temp,counters.hier_direct)() );
  if ( temp != counters.hier_direct )
    printf( "%-40s %s\n", "SHOULD-BE", 
	    valueBody(counters.hier_direct,counters.hier_direct)() );
  puts("");
}

void
showPeaks( const Counters& counters )
{
  char tbuf[256];

  printf( "%-19s %8s %8s      %8s %8s      %8s      %8s      %8s\n",
	  "# REQUEST PEAKS", "udp", "udp-hit", "tcp", "tcp-hit",
	  "direct", "parent", "peer" );
  puts( "------------------- -------- ------------- -------- ------------- ------------- ------------- -------------" );
  for ( MyUInt32 i=globals.startTime / globals.peakInterval;
	i <= globals.stopTime / globals.peakInterval; 
	++i ) {
    time_t then = i * globals.peakInterval;
    strftime( tbuf, sizeof(tbuf), timeFormat, localtime(&then) );
    printf( "%s " SF_wU32(8) " " SF_wU32(8) "%s" SF_wU32(8) " "
	    SF_wU32(8) "%s" SF_wU32(8) "%s" SF_wU32(8) "%s" SF_wU32(8) "%s\n",
	    tbuf,
	    counters.peak_udp[i].reqs(), 
	    counters.peak_udp_hit[i].reqs(),
	    autoPercent( counters.peak_udp[i].reqs(),
			 counters.peak_udp_hit[i].reqs() )(),
	    counters.peak_tcp[i].reqs(), 
	    counters.peak_tcp_hit[i].reqs(),
	    autoPercent( counters.peak_tcp[i].reqs(),
			 counters.peak_tcp_hit[i].reqs() )(),
	    counters.peak_hier_direct[i].reqs(),
	    autoPercent( counters.peak_tcp[i].reqs(), 
			 counters.peak_hier_direct[i].reqs() )(),
	    counters.peak_hier_parent[i].reqs(),
	    autoPercent( counters.peak_tcp[i].reqs(), 
			 counters.peak_hier_parent[i].reqs() )(),
	    counters.peak_hier_peer[i].reqs(),
	    autoPercent( counters.peak_tcp[i].reqs(), 
			 counters.peak_hier_peer[i].reqs() )() );
  }
  puts("");

  printf( "%-19s %10s %10s      %10s %10s      %10s      %10s      %10s\n",
	  "# VOLUME PEAKS", "udp", "udp-hit", "tcp", "tcp-hit",
	  "direct", "parent", "peer" );
  puts( "------------------- ---------- --------------- ---------- --------------- --------------- --------------- ---------------" );
  for ( MyUInt32 i=globals.startTime / globals.peakInterval;
	i <= globals.stopTime / globals.peakInterval; 
	++i ) {
    time_t then = i * globals.peakInterval;
    strftime( tbuf, sizeof(tbuf), timeFormat, localtime(&then) );
    printf( "%s %s %s%s%s %s%s%s%s%s%s%s%s\n",
	    tbuf,
	    autoSize( counters.peak_udp[i].size() )(), 
	    autoSize( counters.peak_udp_hit[i].size() )(),
	    autoPercent( counters.peak_udp[i].size(),
			 counters.peak_udp_hit[i].size() )(),
	    autoSize( counters.peak_tcp[i].size() )(), 
	    autoSize( counters.peak_tcp_hit[i].size() )(),
	    autoPercent( counters.peak_tcp[i].size(),
			 counters.peak_tcp_hit[i].size() )(),
	    autoSize( counters.peak_hier_direct[i].size() )(),
	    autoPercent( counters.peak_tcp[i].size(), 
			 counters.peak_hier_direct[i].size() )(),
	    autoSize( counters.peak_hier_parent[i].size() )(),
	    autoPercent( counters.peak_tcp[i].size(), 
			 counters.peak_hier_parent[i].size() )(),
	    autoSize( counters.peak_hier_peer[i].size() )(),
	    autoPercent( counters.peak_tcp[i].size(), 
			 counters.peak_hier_peer[i].size() )() );
  }
  puts("");
}

static
String
sftime( double stamp )
{
  char temp[3][32];
  double integer;
  sprintf( temp[0], "%.3f", modf( stamp, &integer ) );
  time_t what = integer;
  
  strftime( temp[1], 32, "%H:%M:%S", localtime(&what) );
  strcpy( temp[2], temp[1] );
  strcat( temp[2], temp[0]+1 );
  return String(temp[2]);
}

void 
showStats( DNSCache& dns, IRRCache* irr,
	   double looptime, double startup, MyUInt32 lineno )
{
  double runtime = now();
  puts( "# STATISTICS" );
  printf( "analysis started at %s\n", sftime(startup)() );
  printf( "parsing finished at %s\n", sftime(startup+looptime)() );
  printf( "analysis finished at %s\n", sftime(runtime)() );

  runtime -= startup;
  printf( "%.1f s spent in analyser\n", runtime );
  printf( "%.1f s spent in parser loop\n"
	  SF_U32 " lines processed (%.1f lps)\n", 
	  looptime, lineno, lineno / looptime );
  
  printf( "DNS cache had " SF_U32 " requests with " SF_U32 " hits (%s)\n",
	  dns.request(BaseCache::C_ALL), dns.request(BaseCache::C_HIT),
	  autoPercent( dns.request(BaseCache::C_ALL), 
		       dns.request(BaseCache::C_HIT) ).trim()() );
  if ( irr )
    printf( "IRR cache had " SF_U32 " requests with " SF_U32 " hits (%s)\n",
	    irr->request(BaseCache::C_ALL), irr->request(BaseCache::C_HIT),
	    autoPercent( irr->request(BaseCache::C_ALL), 
			 irr->request(BaseCache::C_HIT) ).trim()() );
  puts("");
}
