//
// $Id: regstr.cc,v 1.3 1998/06/22 12:06:46 voeckler Exp $
//
// Author:  Jens-S. Vckler <voeckler@rvs.uni-hannover.de>
//
// File:    regstr.cc
//          Sat Aug  9 1997
//
// (c) 1997 Lehrgebiet Rechnernetze und Verteilte Systeme
//          Universitt Hannover, Germany
//
// Permission to use, copy, modify, distribute, and sell this software
// and its documentation for any purpose is hereby granted without fee,
// provided that (i) the above copyright notices and this permission
// notice appear in all copies of the software and related documentation,
// and (ii) the names of the Lehrgebiet Rechnernetze und Verteilte
// Systeme and the University of Hannover may not be used in any
// advertising or publicity relating to the software without the
// specific, prior written permission of Lehrgebiet Rechnernetze und
// Verteilte Systeme and the University of Hannover.
//
// THE SOFTWARE IS PROVIDED "AS-IS" AND WITHOUT WARRANTY OF ANY KIND,
// EXPRESS, IMPLIED OR OTHERWISE, INCLUDING WITHOUT LIMITATION, ANY
// WARRANTY OF MERCHANTABILITY OR FITNESS FOR A PARTICULAR PURPOSE.
//
// IN NO EVENT SHALL THE LEHRGEBIET RECHNERNETZE UND VERTEILTE SYSTEME OR
// THE UNIVERSITY OF HANNOVER BE LIABLE FOR ANY SPECIAL, INCIDENTAL,
// INDIRECT OR CONSEQUENTIAL DAMAGES OF ANY KIND, OR ANY DAMAGES
// WHATSOEVER RESULTING FROM LOSS OF USE, DATA OR PROFITS, WHETHER OR NOT
// ADVISED OF THE POSSIBILITY OF DAMAGE, AND ON ANY THEORY OF LIABILITY,
// ARISING OUT OF OR IN CONNECTION WITH THE USE OR PERFORMANCE OF THIS
// SOFTWARE.
//
// $Log: regstr.cc,v $
// Revision 1.3  1998/06/22 12:06:46  voeckler
// added match and special case of gmatch, added another flavour of gmatch.
//
// Revision 1.2  1997/08/18 11:49:09  voeckler
// extended the number of subexpressions to 10, new function gmatch
// allows to extract all subexpressions.
//
// Revision 1.1  1997/08/11 10:12:05  voeckler
// Initial revision
//

#if defined(__GNUC__) || defined(__GNUG__)
#pragma implementation
#endif

#include <string.h>
#include "string.hh"
#include "regstr.hh"

extern "C" {
#include <regex.h>
}

#if defined(SUNOS) && !defined(SOLARIS) || defined(SINIXY)
#include "sundefs.h"
#endif // old SUNOS

static const char* RCS_ID = 
"$Id: regstr.cc,v 1.3 1998/06/22 12:06:46 voeckler Exp $";

RegExp::RegExp( const char* re, int flags )
{
  status = regcomp( &_expression, re, flags );
  _expression.re_nsub = 10;
}

RegExp::RegExp( const String& re, int flags )
{
  status = regcomp( &_expression, re(), flags );
  _expression.re_nsub = 10;
}

RegExp::~RegExp()
{
  regfree( &_expression );
}

String
RegExp::error( void ) const
{
  char buffer[256];
  regerror( status, &_expression, buffer, 256 );
  return String(buffer);
}

bool
RegExp::match( const char* str, unsigned start, int flags )
{
  status = regexec( &_expression, str+start, 10, _match, flags );
  return ( status == 0 );
}

bool
RegExp::match( const String& s, unsigned start, int flags )
{
  return match( s(), start, flags );
}

String 
RegExp::matches( const char* str, unsigned start, int flags )
{
  status = regexec( &_expression, str+start, 10, _match, flags );
  if ( status ) return String();
  else {
    unsigned size = _match[0].rm_eo - _match[0].rm_so;
    return String( str, size, start + _match[0].rm_so );
  }
}

String 
RegExp::matches( const String& s, unsigned start, int flags )
{
  return matches( s(), start, flags );
}

int
RegExp::gmatch( Matches& result, const char* s, unsigned start, int flags )
{
  int count = -1;

  if ( (status=regexec( &_expression, s+start, 10, _match, flags )) == 0 ) 
    for ( int i=0; i<10; i++ ) {
      unsigned size = _match[i].rm_eo - _match[i].rm_so;
      if ( size ) {
	count++;
	result[i] = substring( s, start + _match[i].rm_so, size );
      }
    }

  return count;
}

int
RegExp::gmatch( Matches& result, const String& s, unsigned start, int flags )
{
  int count = -1;

  if ( (status=regexec( &_expression, s()+start, 10, _match, flags )) == 0 ) 
    for ( int i=0; i<10; i++ ) {
      unsigned size = _match[i].rm_eo - _match[i].rm_so;
      if ( size ) {
	count++;
	result[i] = s.substring( start + _match[i].rm_so, size );
      }
    }

  return count;
}

String
RegExp::gmatch( const char* s, unsigned subexp, unsigned start, int flags )
{
  if ( subexp < 10 &&
       (status=regexec( &_expression, s+start, 10, _match, flags )) == 0 ) {
    unsigned size = _match[subexp].rm_eo - _match[subexp].rm_so;
    return substring( s, start + _match[subexp].rm_so, size );
  } else {
    return String();
  }
}

String
RegExp::gmatch( const String& s, unsigned subexp, unsigned start, int flags )
{
  if ( subexp < 10 &&
       (status=regexec( &_expression, s()+start, 10, _match, flags )) == 0 ) {
    unsigned size = _match[subexp].rm_eo - _match[subexp].rm_so;
    return s.substring( start + _match[subexp].rm_so, size );
  } else {
    return String();
  }
}

