//
// $Id: regstr.hh,v 1.3 1998/06/22 12:06:46 voeckler Exp $
//
// Author:  Jens-S. Vckler <voeckler@rvs.uni-hannover.de>
//
// File:    regstr.hh
//          Sat Aug  9 1997
//
// (c) 1997 Lehrgebiet Rechnernetze und Verteilte Systeme
//          Universitt Hannover, Germany
//
// Permission to use, copy, modify, distribute, and sell this software
// and its documentation for any purpose is hereby granted without fee,
// provided that (i) the above copyright notices and this permission
// notice appear in all copies of the software and related documentation,
// and (ii) the names of the Lehrgebiet Rechnernetze und Verteilte
// Systeme and the University of Hannover may not be used in any
// advertising or publicity relating to the software without the
// specific, prior written permission of Lehrgebiet Rechnernetze und
// Verteilte Systeme and the University of Hannover.
//
// THE SOFTWARE IS PROVIDED "AS-IS" AND WITHOUT WARRANTY OF ANY KIND,
// EXPRESS, IMPLIED OR OTHERWISE, INCLUDING WITHOUT LIMITATION, ANY
// WARRANTY OF MERCHANTABILITY OR FITNESS FOR A PARTICULAR PURPOSE.
//
// IN NO EVENT SHALL THE LEHRGEBIET RECHNERNETZE UND VERTEILTE SYSTEME OR
// THE UNIVERSITY OF HANNOVER BE LIABLE FOR ANY SPECIAL, INCIDENTAL,
// INDIRECT OR CONSEQUENTIAL DAMAGES OF ANY KIND, OR ANY DAMAGES
// WHATSOEVER RESULTING FROM LOSS OF USE, DATA OR PROFITS, WHETHER OR NOT
// ADVISED OF THE POSSIBILITY OF DAMAGE, AND ON ANY THEORY OF LIABILITY,
// ARISING OUT OF OR IN CONNECTION WITH THE USE OR PERFORMANCE OF THIS
// SOFTWARE.
//
//
// $Log: regstr.hh,v $
// Revision 1.3  1998/06/22 12:06:46  voeckler
// added match and special case of gmatch, added another flavour of gmatch.
//
// Revision 1.2  1997/08/18 11:49:09  voeckler
// extended the number of subexpressions to 10, new function gmatch
// allows to extract all subexpressions.
//
// Revision 1.1  1997/08/11 10:12:05  voeckler
// Initial revision
//

#ifndef _REGSTR_HH
#define _REGSTR_HH

#if defined(__GNUC__) || defined(__GNUG__)
#pragma interface
#else
#ifndef HAS_BOOL
#define HAS_BOOL
typedef int bool;
#define false 0
#define true  1
#endif
#endif

#include <regex.h>

class String;
typedef String Matches[10];

class RegExp {
public:
  // init+compile and destroy
  RegExp( const char* re, int flags = REG_EXTENDED );
  // purpose: 
  RegExp( const String& re, int flags = REG_EXTENDED );
  ~RegExp();

  inline int lastStatus( void ) const { return status; }
  // just check, if the string matches the re, return true or false.
  // use, if ctor flag REG_NOSUB was used...
  bool match( const String& s, unsigned start = 0, int flags = 0 );
  bool match( const char* str, unsigned start = 0, int flags = 0 );
  // return the matching content
  // don't use, if REG_NOSUB was set in ctor...
  String matches( const String& s, unsigned start = 0, int flags = 0 );
  String matches( const char* str, unsigned start = 0, int flags = 0 );

  int gmatch( Matches& result, 
	      const char* s, unsigned start = 0, int flags = 0 );
  int gmatch( Matches& result, 
	      const String& s, unsigned start = 0, int flags = 0 );
  // purpose: return the number of submatches, and the first 9 submatches
  // paramtr: result (OUT): complete match and first nine submatches
  //          s (IN): string to search for re
  //          start (IN): offset into string
  //          flags (IN): flags, see regexec(3)
  // returns: number of matched expression, 0 for no match.
  // comment: result[0] contains the complete match

  String gmatch( const char* s, unsigned subexp,
		 unsigned start = 0, int flags = 0 );
  String gmatch( const String& s, unsigned subexp,
		 unsigned start = 0, int flags = 0 );
  // purpose: return the nth (maximum 9th) matching subexpression
  // paramtr: s (IN): string to search within for the re
  //          subexp (IN): index of the nth paranthesis to return
  //          start (IN): start offset within string
  //          flags (IN): flags, see regexec(3)
  // returns: matching subexpression, or empty string on no match

  String error( void ) const;

protected:
  // yield error for these -- you *must* *not* activate the default
  // ctor without changing the code...
  RegExp();
  RegExp( const RegExp& );
  RegExp& operator=( const RegExp& );

private:
  int status;
  regex_t    _expression;
  regmatch_t _match[10];
};

#endif // _REGSTR_HH
